# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.

from typing import Sequence, Optional
import base64
import re


def replace_escapes(args: str) -> str:
    """Replace special characters in a given string.

    Parameters:
    - s (str): The input string to be processed.

    Returns:
    - str: The string with the special characters replaced.
    """
    replacements = {
        # Apply the backslash characters rules from CommandLineToArgvW
        # https://learn.microsoft.com/en-us/archive/blogs/twistylittlepassagesallalike/everyone-quotes-command-line-arguments-the-wrong-way
        "\\": "\\\\",
        # This is a known issue in PowerShell 5 without installing additional `Native` module
        # https://github.com/MicrosoftDocs/PowerShell-Docs/issues/2361
        '"': '\\"',
        # To include a single quotation mark in a single-quoted string, use a second consecutive single quote.
        # https://learn.microsoft.com/en-us/powershell/module/microsoft.powershell.core/about/about_quoting_rules
        "'": "''",
    }
    return re.sub(r'[\'\\"]', lambda m: replacements[m.group(0)], args)


def generate_exit_code_powershell(cmd_line: str) -> str:
    """Generate a PowerShell script string that runs the provided command and then exits with
    the command's exit code. This can be useful when needing to propagate the exit code of a
    PowerShell Script subprocess to the parent process.

    Args:
        cmd_line (str): The command line string that needs to be executed in the generated PowerShell script.

    Returns:
        str: A PowerShell script string that executes the provided command and then exits with its exit code.
    """
    return f"""
$process = {cmd_line}
exit $process.ExitCode
"""


def generate_exit_code_wrapper(args: Sequence[str], exit_action: Optional[str] = None) -> str:
    """Generate a wrapper outside the command used for catching the exit code and return it correctly.

    If the command is not found or the exit code is $null, exit code will be returned as 1.

    Args:
        args: command arguments passed by users.
        special_character_replace_fn: A function used for replacing particular characters in the command.
        exit_action: A string inserted to the wrapper to determine what the action is when the task is done.
            If it is None, the script will exit with the exit code.

    Returns:
        str: A PowerShell Script contains the user's commands with the exit code wrapper.
    """
    quoted_args = [f"'{replace_escapes(arg)}'" for arg in args]
    exit_action = exit_action if exit_action else "exit"
    cmd_line = " ".join(quoted_args)
    return f"""
try {{
    # Attempt to run the command. This will fail if the command is not recognized.
    & {cmd_line}
}}
catch [System.Management.Automation.CommandNotFoundException] {{
    Write-Host "Command not found: $_, exiting with code 1."
    {exit_action} 1
}}
catch {{
    Write-Host "An unexpected error occurred: $_"
    if ($LASTEXITCODE -eq $null) {{
        Write-Host "The original exit code is null. Exit with code = 1"
        {exit_action} 1  # Exit with code 1 if command does not set an exit code
    }}
    {exit_action} $LASTEXITCODE
}}

{exit_action} $LASTEXITCODE
"""


def encode_to_base64(command: str) -> str:
    """Encodes the provided string command to a base64 string using UTF-16LE encoding.
    This method is useful when needing an encoding equivalent to PowerShell's Unicode encoding.

    Args:
        command (str): The original command string needed to be encoded.

    Returns:
        str: The base64 encoded version of the input command using UTF-16LE encoding.
    """
    command_bytes = command.encode("utf-16le")
    encoded_command = base64.b64encode(command_bytes)
    return encoded_command.decode("utf-8")



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.

from ._os_checker import is_windows

if is_windows():
    import win32security
    import ntsecuritycon


class WindowsPermissionHelper:
    """
    This class contains helper methods to set permissions for files and directories on Windows.
    """

    @staticmethod
    def set_permissions_full_control(file_path, principals_to_permit):
        """
        Grants full control over the object at file_path to all principals in principals_to_permit.
        Sets flags so both child files and directories inherit these permissions.

        Arguments:
            file_path (str): The path to the file or directory.
            principals_to_permit (List[str]): The names of the principals to permit.

        Raises:
            RuntimeError if there is a problem modifying the security attributes.
        """
        try:
            # We don't want to propagate existing permissions, so create a new DACL
            dacl = win32security.ACL()
            for principal in principals_to_permit:
                user_or_group_sid, _, _ = win32security.LookupAccountName(None, principal)

                # Add an ACE to the DACL giving the principal full control and enabling inheritance of the ACE
                dacl.AddAccessAllowedAceEx(
                    win32security.ACL_REVISION,
                    ntsecuritycon.OBJECT_INHERIT_ACE | ntsecuritycon.CONTAINER_INHERIT_ACE,
                    ntsecuritycon.FILE_ALL_ACCESS,
                    user_or_group_sid,
                )

            # Get the security descriptor of the tempdir
            sd = win32security.GetFileSecurity(
                str(file_path), win32security.DACL_SECURITY_INFORMATION
            )

            # Set the security descriptor's DACL to the newly-created DACL
            # Arguments:
            # 1. bDaclPresent = 1: Indicates that the DACL is present in the security descriptor.
            #    If set to 0, this method ignores the provided DACL and allows access to all principals.
            # 2. dacl: The discretionary access control list (DACL) to be set in the security descriptor.
            # 3. bDaclDefaulted = 0: Indicates the DACL was provided and not defaulted.
            #    If set to 1, indicates the DACL was defaulted, as in the case of permissions inherited from a parent directory.
            sd.SetSecurityDescriptorDacl(1, dacl, 0)

            # Set the security descriptor to the tempdir
            win32security.SetFileSecurity(
                str(file_path), win32security.DACL_SECURITY_INFORMATION, sd
            )
        except Exception as err:
            raise RuntimeError(
                f"Could not change permissions of directory '{str(dir)}' (error: {str(err)})"
            )



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.

# Note: This script will only be run as a subprocess.

import sys
from psutil import Process
from signal import CTRL_BREAK_EVENT  # type: ignore
from win32console import AttachConsole, FreeConsole  # type: ignore


def send_ctrl_break_event(pgid: int):
    """Sends a CTRL_BREAK_EVENT to a process group id to signal process to shut down"""
    process = Process(pgid)

    # Send signal can only target processes in the same console.
    # We first detach from the current console and re-attach to that of process group.
    FreeConsole()
    AttachConsole(pgid)

    # Send the signal
    # We send CTRL-BREAK as handler for it cannnot be disabled.
    # https://learn.microsoft.com/en-us/windows/console/ctrl-c-and-ctrl-break-signals
    process.send_signal(CTRL_BREAK_EVENT)


pgid = int(sys.argv[1])
send_ctrl_break_event(pgid)



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.

import sys
from ._os_checker import is_windows
from typing import Any

if is_windows():
    import ctypes
    import win32process
    from ctypes import wintypes
    from subprocess import Handle, list2cmdline, Popen  # type: ignore

# Tell type checker to ignore on non-windows platforms
assert sys.platform == "win32"

advapi32 = ctypes.WinDLL("advapi32")
kernel32 = ctypes.WinDLL("kernel32")

# Constants
LOGON_WITH_PROFILE = 0x00000001


# Structures
# https://learn.microsoft.com/en-us/windows/win32/api/processthreadsapi/ns-processthreadsapi-startupinfoa
class STARTUPINFO(ctypes.Structure):
    _fields_ = [
        ("cb", wintypes.DWORD),
        ("lpReserved", wintypes.LPWSTR),
        ("lpDesktop", wintypes.LPWSTR),
        ("lpTitle", wintypes.LPWSTR),
        ("dwX", wintypes.DWORD),
        ("dwY", wintypes.DWORD),
        ("dwXSize", wintypes.DWORD),
        ("dwYSize", wintypes.DWORD),
        ("dwXCountChars", wintypes.DWORD),
        ("dwYCountChars", wintypes.DWORD),
        ("dwFillAttribute", wintypes.DWORD),
        ("dwFlags", wintypes.DWORD),
        ("wShowWindow", wintypes.WORD),
        ("cbReserved2", wintypes.WORD),
        ("lpReserved2", ctypes.POINTER(wintypes.BYTE)),
        ("hStdInput", wintypes.HANDLE),
        ("hStdOutput", wintypes.HANDLE),
        ("hStdError", wintypes.HANDLE),
    ]


# https://learn.microsoft.com/en-us/windows/win32/api/processthreadsapi/ns-processthreadsapi-process_information
class PROCESS_INFORMATION(ctypes.Structure):
    _fields_ = [
        ("hProcess", wintypes.HANDLE),
        ("hThread", wintypes.HANDLE),
        ("dwProcessId", wintypes.DWORD),
        ("dwThreadId", wintypes.DWORD),
    ]


class PopenWindowsAsUser(Popen):
    """Class to run a process as another user on Windows.
    Derived from Popen, it defines the _execute_child() method to call CreateProcessWithLogonW.
    """

    def __init__(self, username: str, password: str, *args: Any, **kwargs: Any):
        """
        Arguments:
            username (str):  Name of user to run subprocess as
            password (str):  Password for username
            args (Any):  Popen constructor args
            kwargs (Any):  Popen constructor kwargs
            https://docs.python.org/3/library/subprocess.html#popen-constructor
        """
        self.username = username
        self.password = password
        self.domain = None
        super(PopenWindowsAsUser, self).__init__(*args, **kwargs)

    def _execute_child(
        self,
        args,
        executable,
        preexec_fn,
        close_fds,
        pass_fds,
        cwd,
        env,
        startupinfo,
        creationflags,
        shell,
        p2cread,
        p2cwrite,
        c2pread,
        c2pwrite,
        errread,
        errwrite,
        restore_signals,
        start_new_session,
        *additional_args,
        **kwargs,
    ):
        """Execute program (MS Windows version).
        Calls CreateProcessWithLogonW to run a process as another user.
        """

        assert not pass_fds, "pass_fds not supported on Windows."

        commandline = args if isinstance(args, str) else list2cmdline(args)

        # Initialize structures
        si = STARTUPINFO()
        si.cb = ctypes.sizeof(STARTUPINFO)
        pi = PROCESS_INFORMATION()

        si.hStdInput = int(p2cread)
        si.hStdOutput = int(c2pwrite)
        si.hStdError = int(errwrite)
        si.dwFlags |= win32process.STARTF_USESTDHANDLES

        # https://learn.microsoft.com/en-us/windows/win32/api/winbase/nf-winbase-createprocesswithlogonw
        result = advapi32.CreateProcessWithLogonW(
            self.username,
            self.domain,
            self.password,
            LOGON_WITH_PROFILE,
            executable,
            commandline,
            creationflags,
            env,
            cwd,
            ctypes.byref(si),
            ctypes.byref(pi),
        )

        if not result:
            raise ctypes.WinError()

        # Child is launched. Close the parent's copy of those pipe
        # handles that only the child should have open.
        self._close_pipe_fds(p2cread, p2cwrite, c2pread, c2pwrite, errread, errwrite)

        # Retain the process handle, but close the thread handle
        kernel32.CloseHandle(pi.hThread)

        self._child_created = True
        self.pid = pi.dwProcessId
        self._handle = Handle(pi.hProcess)



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.

import os
import re

from ._os_checker import is_posix, is_windows

if is_posix():
    import grp
    import pwd


if is_windows():
    import win32api
    import win32security
    import win32net
    import win32netcon
    import win32con
    import pywintypes
    import winerror
    from win32con import LOGON32_LOGON_INTERACTIVE, LOGON32_PROVIDER_DEFAULT

from typing import Optional, Tuple, Union

from abc import ABC, abstractmethod

__all__ = (
    "PosixSessionUser",
    "SessionUser",
    "WindowsSessionUser",
    "BadCredentialsException",
    "BadUserNameException",
)


class BadCredentialsException(Exception):
    """Exception raised for incorrect username or password."""

    pass


class BadUserNameException(Exception):
    """Exception raised for incorrect username."""

    pass


class BadDomainNameException(Exception):
    """Exception raised for incorrect Domain Name."""

    pass


class SessionUser(ABC):
    """Base class for holding information on the specific os-user identity to run
    a Session as.
    """

    user: str
    """
    User name of the identity to run the Session's subprocesses under.
    """

    @staticmethod
    @abstractmethod
    def get_process_user():
        """
        Returns the user name of the user running the current process.
        """
        pass

    def is_process_user(self) -> bool:
        """
        Returns True if the session user is the user running the current process, else False.

        """
        return self.user == self.get_process_user()


class PosixSessionUser(SessionUser):
    __slots__ = ("user", "group")
    """Specific os-user identity to run a Session as under Linux/macOS."""

    user: str
    """User name of the identity to run the Session's subprocesses under.
    """

    group: str
    """Group name of the identity to run the Session's subprocesses under.
    """

    def __init__(self, user: str, *, group: Optional[str] = None) -> None:
        """
        Arguments:
            user (str): The user
            group (Optional[str]): The group. Defaults to the name of this
                process' effective group.
        """
        if not is_posix():
            raise RuntimeError("Only available on posix systems.")
        self.user = user
        self.group = group if group else grp.getgrgid(os.getegid()).gr_name  # type: ignore

    @staticmethod
    def get_process_user():
        """
        Returns the user name of the user running the current process.
        """
        return pwd.getpwuid(os.geteuid()).pw_name


class WindowsSessionUser(SessionUser):
    __slots__ = ("user", "group", "password")
    """Specific os-user identity to run a Session as under Windows."""

    user: str
    """
    User name of the identity to run the Session's subprocesses under.
    This can be either a plain username for a local user or a domain username in down-level logon form
    ex: localUser, domain\\domainUser
    """

    group: Optional[str]
    """
    Group name of the identity to run the Session's subprocesses under.
    This can be just a group name for a local group, or a domain group in down-level logon form.
    ex: localGroup, domain\\domainGroup
    """

    password: Optional[str]
    """
    Password of the identity to run the Session's subprocess under.
    """

    def __init__(
        self, user: str, *, password: Optional[str] = None, group: Optional[str] = None
    ) -> None:
        """
        Arguments:
            user (str): User name of the identity to run the Session's subprocesses under.
                        This can be either a plain username for a local user, a domain username in down-level logon form,
                        or a domain's UPN.
                        ex: localUser, domain\\domainUser, domainUser@domain.com
            group (Optional[str]): Group name of the identity to run the Session's subprocesses under.
                         This can be just a group name for a local group, or a domain group in down-level format.
                         ex: localGroup, domain\\domainGroup
            password (Optional[str]): Password of the identity to run the Session's subprocess under.
        """
        if not is_windows():
            raise RuntimeError("Only available on Windows systems.")

        self.group = group
        self.password = password

        if "@" in user and self.is_domain_joined():
            user = win32security.TranslateName(
                user, win32api.NameUserPrincipal, win32api.NameSamCompatible
            )

        self.user = user

        domain, username_without_domain = WindowsSessionUser.split_domain_and_username(user)

        self.is_valid_username(username_without_domain)
        if domain:
            self.is_valid_domain(domain)

        if password is None and not self.is_process_user():
            raise RuntimeError(
                "Without passing a password, WindowsSessionUser's user must match the user running Open Job Description."
            )

        if password is not None:
            self.validate_username_password(user, domain, password)

    @staticmethod
    def is_domain_joined() -> bool:
        """
        Returns True if the machine is joined to a domain, else False.
        """
        _, join_status = win32net.NetGetJoinInformation()
        return join_status != win32netcon.NetSetupUnjoined

    @staticmethod
    def get_process_user():
        """
        Returns the user name of the user running the current process.
        """
        return win32api.GetUserNameEx(win32con.NameSamCompatible)

    @staticmethod
    def split_domain_and_username(user_name_with_domain: str) -> Tuple[Optional[str], str]:
        """
        Splits a username with domain into domain and username.

        Args:
            user_name_with_domain:  Username needed to be split.
        Returns:
            tuple[Optional[str], str]: domain and username. domain is None if the username is not a domain username.
        """

        domain = None
        user_name = user_name_with_domain
        if "\\" in user_name_with_domain and WindowsSessionUser.is_domain_joined():
            domain, user_name = user_name_with_domain.split("\\")
        return domain, user_name

    @staticmethod
    def validate_username_password(
        user_name: str, domain_name: Union[str, None], password: str
    ) -> Optional[bool]:
        """
        Validates the username and password against Windows authentication.

        Args:
            user_name (str): The username to be validated.
            domain_name (str): The domain where the user exists. None means current domain.
            password (str): The password to be validated.

        Returns:
            Optional[bool]: True if the credentials are valid

        Raises:
            BadCredentialsException: If the username or password is incorrect.
        """
        try:
            handle = win32security.LogonUser(
                user_name,
                domain_name,
                password,
                LOGON32_LOGON_INTERACTIVE,
                LOGON32_PROVIDER_DEFAULT,
            )
            win32api.CloseHandle(handle)
            return True
        except pywintypes.error as e:
            if e.winerror == winerror.ERROR_LOGON_FAILURE:
                raise BadCredentialsException("The username or password is incorrect.")
            raise

    @staticmethod
    def is_valid_username(username: str) -> Optional[bool]:
        """
        Validates a username based on specific rules.
        Reference:
        https://learn.microsoft.com/en-us/windows-hardware/customize/desktop/unattend/microsoft-windows-shell-setup-autologon-username#values

        This function checks if the given username adheres to the following criteria:
        1. It must be a string.
        2. Its length should not exceed 256 characters.
        3. It should not contain any restricted characters ("/[]:|<>+=;,?*%@).
        4. It should not be the specific name "NONE".

        Parameters:
            username (str): The username to be validated.

        Returns:
            Optional[bool]: True if the username is valid according to the above rules, False otherwise.

        Raises:
            BadUserNameException: If the username doesn't follow the rule.
        """

        if not isinstance(username, str):
            raise BadUserNameException(
                f"Username should be a string, not {type(username).__name__}."
            )

        if len(username) > 256 or len(username) == 0:
            raise BadUserNameException("Username must have a length between 1 and 256 characters.")

        # Set of restricted characters
        restricted_chars = set('"/[]:|<>+=;,?*%@')

        if restricted_chars.intersection(username):
            raise BadUserNameException(
                f"Username contains restricted characters {restricted_chars.intersection(username)}. "
                f"Allowed characters do not include any of {restricted_chars}."
            )

        if username.upper() == "NONE":
            raise BadUserNameException("Username cannot be 'NONE'.")

        return True

    @staticmethod
    def is_valid_domain(domain: str) -> Optional[bool]:
        """
        Validates the domain name based on specific criteria.
        Reference:
        https://learn.microsoft.com/en-us/troubleshoot/windows-server/identity/naming-conventions-for-computer-domain-site-ou#dns-domain-names

        This function checks if a given domain name adheres to certain standards:
        - The domain name must only contain alphabetic characters (A-Z,a-z), numeric characters (0-9),
          the minus sign (-), and the period (.).
        - The domain name's length must be between 2 and 255 characters.

        Args:
            domain (str): The domain name to validate.

        Returns:
            Optional[bool]: Returns True if the domain name is valid.

        Raises:
            BadDomainNameException: If the domain name contains disallowed characters or does not meet the
            length requirements, this exception is raised with a message detailing the validation issue.
        """

        # Regex pattern to match only allowed characters
        pattern = r"^[A-Za-z0-9\-.]+$"

        # Check if the string matches the pattern
        if not re.match(pattern, domain):
            raise BadDomainNameException(
                f"Domain name '{domain}' contains disallowed characters."
                "alphabetic characters (A-Z), numeric characters (0-9), "
                "the minus sign (-), and the period (.) are only allowed in the domain name."
            )

        if len(domain) > 255 or len(domain) < 2:
            raise BadDomainNameException(
                "Domain name must have a length between 2 and 255 characters."
            )

        return True
